from __future__ import absolute_import

import subprocess

from semantic_version import Version

from .exceptions import SolcError
from .install import get_executable
from .utils.string import coerce_return_to_text, force_bytes


@coerce_return_to_text
def solc_wrapper(
    solc_binary=None,
    stdin=None,
    help=None,
    version=None,
    add_std=None,
    combined_json=None,
    optimize=None,
    optimize_runs=None,
    libraries=None,
    output_dir=None,
    gas=None,
    assemble=None,
    link=None,
    source_files=None,
    import_remappings=None,
    ast=None,
    ast_json=None,
    asm=None,
    asm_json=None,
    opcodes=None,
    bin=None,
    bin_runtime=None,
    clone_bin=None,
    abi=None,
    hashes=None,
    userdoc=None,
    devdoc=None,
    formal=None,
    allow_paths=None,
    standard_json=None,
    success_return_code=0,
    evm_version=None,
):
    if solc_binary is None:
        solc_binary = get_executable()

    command = [solc_binary]

    solc_minor = Version(solc_binary.rsplit("-v")[-1].split("\\")[0]).minor

    if help:
        command.append("--help")

    if version:
        command.append("--version")

    # removed in 0.4.21 and does nothing since <0.4.11, should be removed in the future
    if add_std:
        command.append("--add-std")

    if optimize:
        command.append("--optimize")

    if optimize_runs is not None:
        command.extend(("--optimize-runs", str(optimize_runs)))

    if link:
        command.append("--link")

    if libraries is not None:
        command.extend(("--libraries", libraries))

    if output_dir is not None:
        command.extend(("--output-dir", output_dir))

    if combined_json:
        if solc_minor >= 5:
            combined_json = combined_json.replace(",clone-bin", "")
        command.extend(("--combined-json", combined_json))

    if gas:
        command.append("--gas")

    if allow_paths:
        command.extend(("--allow-paths", allow_paths))

    if standard_json:
        command.append("--standard-json")

    if assemble:
        command.append("--assemble")

    if import_remappings is not None:
        command.extend(import_remappings)

    if source_files is not None:
        command.extend(source_files)

    # Output configuration
    if ast_json:
        command.append("--ast-json")

    if asm:
        command.append("--asm")

    if asm_json:
        command.append("--asm-json")

    if opcodes:
        command.append("--opcodes")

    if bin:
        command.append("--bin")

    if bin_runtime:
        command.append("--bin-runtime")

    if abi:
        command.append("--abi")

    if hashes:
        command.append("--hashes")

    if userdoc:
        command.append("--userdoc")

    if devdoc:
        command.append("--devdoc")

    if evm_version:
        command.extend(("--evm-version", evm_version))

    # unsupported by >=0.6.0
    if ast:
        if solc_minor >= 6:
            raise AttributeError("solc 0.{}.x does not support the --ast flag".format(solc_minor))
        command.append("--ast")

    # unsupported by >=0.5.0
    if clone_bin:
        if solc_minor >= 5:
            raise AttributeError(
                "solc 0.{}.x does not support the --clone-bin flag".format(solc_minor)
            )
        command.append("--clone-bin")

    if formal:
        if solc_minor >= 5:
            raise AttributeError(
                "solc 0.{}.x does not support the --formal flag".format(solc_minor)
            )
        command.append("--formal")

    if not standard_json and not source_files and solc_minor >= 5:
        command.append("-")

    if stdin is not None:
        # solc seems to expects utf-8 from stdin:
        # see Scanner class in Solidity source
        stdin = force_bytes(stdin, "utf8")

    proc = subprocess.Popen(
        command, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )

    stdoutdata, stderrdata = proc.communicate(stdin)

    if proc.returncode != success_return_code:
        raise SolcError(
            command=command,
            return_code=proc.returncode,
            stdin_data=stdin,
            stdout_data=stdoutdata,
            stderr_data=stderrdata,
        )

    return stdoutdata, stderrdata, command, proc



import numbers
from collections import abc


def is_integer(value):
    return isinstance(value, int) and not isinstance(value, bool)


def is_bytes(value):
    return isinstance(value, (bytes, bytearray))


def is_text(value):
    return isinstance(value, str)


def is_string(value):
    return isinstance(value, (bytes, str, bytearray))


def is_boolean(value):
    return isinstance(value, bool)


def is_dict(obj):
    return isinstance(obj, abc.Mapping)


def is_list_like(obj):
    return not is_string(obj) and isinstance(obj, abc.Sequence)


def is_null(obj):
    return obj is None


def is_number(obj):
    return isinstance(obj, numbers.Number)



import textwrap

from .utils.string import force_text


def force_text_maybe(value, encoding="iso-8859-1"):
    if value is not None:
        return force_text(value)


DEFAULT_MESSAGE = "An error occurred during execution"


class SolcError(Exception):
    message = DEFAULT_MESSAGE

    def __init__(self, command, return_code, stdin_data, stdout_data, stderr_data, message=None):
        if message is not None:
            self.message = message
        self.command = command
        self.return_code = return_code
        self.stdin_data = force_text_maybe(stdin_data, "utf8")
        self.stderr_data = force_text_maybe(stderr_data, "utf8")
        self.stdout_data = force_text_maybe(stdout_data, "utf8")

    def __str__(self):
        return textwrap.dedent(
            (
                """
        {s.message}
        > command: `{command}`
        > return code: `{s.return_code}`
        > stderr:
        {s.stdout_data}
        > stdout:
        {s.stderr_data}
        """
            ).format(
                s=self, command=" ".join(self.command),
            )
        ).strip()


class ContractsNotFound(SolcError):
    message = "No contracts found during compilation"


class SolcNotInstalled(Exception):
    pass


class DownloadError(Exception):
    pass



import os
import sys
import tempfile
import threading
from pathlib import Path

if sys.platform == "win32":
    import msvcrt

    OPEN_MODE = os.O_RDWR | os.O_CREAT | os.O_TRUNC
else:
    import fcntl

    NON_BLOCKING = fcntl.LOCK_EX | fcntl.LOCK_NB
    BLOCKING = fcntl.LOCK_EX

_locks = {}
_base_lock = threading.Lock()


def get_process_lock(lock_id):
    with _base_lock:
        if lock_id not in _locks:
            if sys.platform == "win32":
                _locks[lock_id] = WindowsLock(lock_id)
            else:
                _locks[lock_id] = UnixLock(lock_id)
        return _locks[lock_id]


class _ProcessLock:
    def __init__(self, lock_id):
        self._lock = threading.Lock()
        self._lock_path = Path(tempfile.gettempdir()).joinpath(f".solcx-lock-{lock_id}")
        self._lock_file = self._lock_path.open("w")

    def wait(self):
        self.acquire(True)
        self.release()


class UnixLock(_ProcessLock):
    def acquire(self, blocking):
        if not self._lock.acquire(blocking):
            return False
        try:
            fcntl.flock(self._lock_file, BLOCKING if blocking else NON_BLOCKING)
        except BlockingIOError:
            self._lock.release()
            return False
        return True

    def release(self):
        fcntl.flock(self._lock_file, fcntl.LOCK_UN)
        self._lock.release()


class WindowsLock(_ProcessLock):
    def acquire(self, blocking):
        if not self._lock.acquire(blocking):
            return False
        while True:
            try:
                fd = os.open(self._lock_path, OPEN_MODE)
                msvcrt.locking(fd, msvcrt.LK_LOCK if blocking else msvcrt.LK_NBLCK, 1)
                self._fd = fd
                return True
            except OSError:
                if not blocking:
                    self._lock.release()
                    return False

    def release(self):
        msvcrt.locking(self._fd, msvcrt.LK_UNLCK, 1)
        self._lock.release()



import codecs
import functools

from .types import is_bytes, is_dict, is_list_like, is_string, is_text


def force_bytes(value, encoding="iso-8859-1"):
    if is_bytes(value):
        return bytes(value)
    elif is_text(value):
        return codecs.encode(value, encoding)
    else:
        raise TypeError("Unsupported type: {0}".format(type(value)))


def force_text(value, encoding="iso-8859-1"):
    if is_text(value):
        return value
    elif is_bytes(value):
        return codecs.decode(value, encoding)
    else:
        raise TypeError("Unsupported type: {0}".format(type(value)))


def force_obj_to_bytes(obj):
    if is_string(obj):
        return force_bytes(obj)
    elif is_dict(obj):
        return {k: force_obj_to_bytes(v) for k, v in obj.items()}
    elif is_list_like(obj):
        return type(obj)(force_obj_to_bytes(v) for v in obj)
    else:
        return obj


def force_obj_to_text(obj):
    if is_string(obj):
        return force_text(obj)
    elif is_dict(obj):
        return {k: force_obj_to_text(v) for k, v in obj.items()}
    elif is_list_like(obj):
        return type(obj)(force_obj_to_text(v) for v in obj)
    else:
        return obj


def coerce_args_to_bytes(fn):
    @functools.wraps(fn)
    def inner(*args, **kwargs):
        bytes_args = force_obj_to_bytes(args)
        bytes_kwargs = force_obj_to_bytes(kwargs)
        return fn(*bytes_args, **bytes_kwargs)

    return inner


def coerce_args_to_text(fn):
    @functools.wraps(fn)
    def inner(*args, **kwargs):
        text_args = force_obj_to_text(args)
        text_kwargs = force_obj_to_text(kwargs)
        return fn(*text_args, **text_kwargs)

    return inner


def coerce_return_to_bytes(fn):
    @functools.wraps(fn)
    def inner(*args, **kwargs):
        return force_obj_to_bytes(fn(*args, **kwargs))

    return inner


def coerce_return_to_text(fn):
    @functools.wraps(fn)
    def inner(*args, **kwargs):
        return force_obj_to_text(fn(*args, **kwargs))

    return inner



"""
Install solc
"""
import argparse
import logging
import os
import re
import shutil
import stat
import subprocess
import sys
import tarfile
import tempfile
import zipfile
from base64 import b64encode
from io import BytesIO
from pathlib import Path

import requests
from semantic_version import SimpleSpec, Version

from .exceptions import DownloadError, SolcNotInstalled
from .utils.lock import get_process_lock

try:
    from tqdm import tqdm
except ImportError:
    tqdm = None


DOWNLOAD_BASE = "https://github.com/ethereum/solidity/releases/download/{}/{}"
ALL_RELEASES = "https://api.github.com/repos/ethereum/solidity/releases?per_page=100"

MINIMAL_SOLC_VERSION = "v0.4.11"
VERSION_REGEX = {
    "darwin": "solidity_[0-9].[0-9].[0-9]{1,}.tar.gz",
    "linux": "solc-static-linux",
    "win32": "solidity-windows.zip",
}
LOGGER = logging.getLogger("solcx")

SOLCX_BINARY_PATH_VARIABLE = "SOLCX_BINARY_PATH"

solc_version = None


def _get_platform():
    if sys.platform.startswith("linux"):
        return "linux"
    if sys.platform in ("darwin", "win32"):
        return sys.platform
    raise KeyError(
        "Unknown platform: '{}' - py-solc-x supports" " Linux, OSX and Windows".format(sys.platform)
    )


def get_solc_folder(solcx_binary_path=None):
    if os.getenv(SOLCX_BINARY_PATH_VARIABLE):
        return Path(os.getenv(SOLCX_BINARY_PATH_VARIABLE))
    elif solcx_binary_path is not None:
        return Path(solcx_binary_path)
    else:
        path = Path.home().joinpath(".solcx")
        path.mkdir(exist_ok=True)
        return path


def _import_version(path):
    version = subprocess.check_output([path, "--version"]).decode()
    return "v" + version[version.index("Version: ") + 9 : version.index("+")]


def import_installed_solc(solcx_binary_path=None):
    platform = _get_platform()
    if platform == "win32":
        return

    # copy active version of solc
    path_list = []
    which = subprocess.run(["which", "solc"], stdout=subprocess.PIPE).stdout.decode().strip()
    if which:
        path_list.append(which)

    # on OSX, also copy all versions of solc from cellar
    if platform == "darwin":
        path_list = [str(i) for i in Path("/usr/local/Cellar").glob("solidity*/**/solc")]

    for path in path_list:
        try:
            version = _import_version(path)
            assert version not in get_installed_solc_versions()
        except Exception:
            continue
        copy_path = str(
            get_solc_folder(solcx_binary_path=solcx_binary_path).joinpath("solc-" + version)
        )
        shutil.copy(path, copy_path)
        try:
            # confirm that solc still works after being copied
            assert version == _import_version(copy_path)
        except Exception:
            os.unlink(copy_path)


def get_executable(version=None, solcx_binary_path=None):
    if not version:
        version = solc_version
    if not version:
        raise SolcNotInstalled(
            "Solc is not installed. Call solcx.get_available_solc_versions()"
            " to view for available versions and solcx.install_solc() to install."
        )
    solc_bin = get_solc_folder(solcx_binary_path=solcx_binary_path).joinpath("solc-" + version)
    if sys.platform == "win32":
        solc_bin = solc_bin.joinpath("solc.exe")
    if not solc_bin.exists():
        raise SolcNotInstalled(
            "solc {} has not been installed. ".format(version)
            + "Use solcx.install_solc('{}') to install.".format(version)
        )
    return str(solc_bin)


def set_solc_version(version, silent=False, solcx_binary_path=None):
    version = _check_version(version)
    get_executable(version, solcx_binary_path)
    global solc_version
    solc_version = version
    if not silent:
        LOGGER.info("Using solc version {}".format(solc_version))


def set_solc_version_pragma(pragma_string, silent=False, check_new=False):
    version = _select_pragma_version(
        pragma_string, [Version(i[1:]) for i in get_installed_solc_versions()]
    )
    if not version:
        raise SolcNotInstalled(
            "No compatible solc version installed. "
            + "Use solcx.install_solc_version_pragma('{}') to install.".format(version)
        )
    global solc_version
    solc_version = version
    if not silent:
        LOGGER.info("Using solc version {}".format(solc_version))
    if check_new:
        latest = install_solc_pragma(pragma_string, False)
        if Version(latest) > Version(version):
            LOGGER.info("Newer compatible solc version exists: {}".format(latest))


def install_solc_pragma(pragma_string, install=True, show_progress=False, solcx_binary_path=None):
    version = _select_pragma_version(
        pragma_string, [Version(i[1:]) for i in get_available_solc_versions()]
    )
    if not version:
        raise ValueError("Compatible solc version does not exist")
    if install:
        install_solc(version, show_progress=show_progress, solcx_binary_path=solcx_binary_path)
    return version


def get_available_solc_versions(headers=None):
    versions = []
    pattern = VERSION_REGEX[_get_platform()]

    # Github sometimes blocks CI from calling their API, if you are having issues try
    # saving an API token to the environment variable GITHUB_TOKEN in your build environment
    # https://github.blog/2013-05-16-personal-api-tokens/
    if not headers and os.getenv("GITHUB_TOKEN"):
        auth = b64encode(os.getenv("GITHUB_TOKEN").encode()).decode()
        headers = {"Authorization": "Basic {}".format(auth)}

    data = requests.get(ALL_RELEASES, headers=headers)
    if data.status_code != 200:
        raise ConnectionError(
            "Status {} when getting solc versions from Github: '{}'".format(
                data.status_code, data.json()["message"]
            )
        )

    for release in data.json():
        asset = next((i for i in release["assets"] if re.match(pattern, i["name"])), False)
        if asset:
            versions.append(release["tag_name"])
        if release["tag_name"] == MINIMAL_SOLC_VERSION:
            break
    return versions


def _select_pragma_version(pragma_string, version_list):
    comparator_set_range = pragma_string.replace(" ", "").split("||")
    comparator_regex = re.compile(r"(([<>]?=?|\^)\d+\.\d+\.\d+)+")
    version = None

    for comparator_set in comparator_set_range:
        spec = SimpleSpec(*(i[0] for i in comparator_regex.findall(comparator_set)))
        selected = spec.select(version_list)
        if selected and (not version or version < selected):
            version = selected
    if version:
        return str(version)


def get_installed_solc_versions(solcx_binary_path=None):
    return sorted(
        i.name[5:] for i in get_solc_folder(solcx_binary_path=solcx_binary_path).glob("solc-v*")
    )


def install_solc(version, allow_osx=False, show_progress=False, solcx_binary_path=None):
    platform = _get_platform()
    version = _check_version(version)

    lock = get_process_lock(version)
    if not lock.acquire(False):
        lock.wait()
        if not _check_for_installed_version(version):
            return
        return install_solc(version, allow_osx)

    try:
        if platform == "linux":
            _install_solc_linux(version, show_progress, solcx_binary_path)
        elif platform == "darwin":
            _install_solc_osx(version, allow_osx, show_progress, solcx_binary_path)
        elif platform == "win32":
            _install_solc_windows(version, show_progress, solcx_binary_path)
        binary_path = get_executable(version, solcx_binary_path)
        _check_subprocess_call(
            [binary_path, "--version"],
            message="Checking installed executable version @ {}".format(binary_path),
        )
        if not solc_version:
            set_solc_version(version)
        LOGGER.info("solc {} successfully installed at: {}".format(version, binary_path))
    finally:
        lock.release()


def _check_version(version):
    version = Version(version.lstrip("v"))
    if version not in SimpleSpec(">=0.4.11"):
        raise ValueError("py-solc-x does not support solc versions <0.4.11")
    return "v" + str(version)


def _check_subprocess_call(command, message=None, verbose=False, **proc_kwargs):
    if message:
        LOGGER.debug(message)
    LOGGER.info("Executing: {0}".format(" ".join(command)))

    return subprocess.check_call(
        command, stderr=subprocess.STDOUT if verbose else subprocess.DEVNULL, **proc_kwargs
    )


def _chmod_plus_x(executable_path):
    executable_path.chmod(executable_path.stat().st_mode | stat.S_IEXEC)


def _check_for_installed_version(version, solcx_binary_path=None):
    path = get_solc_folder(solcx_binary_path=solcx_binary_path).joinpath("solc-" + version)
    if path.exists():
        LOGGER.info("solc {} already installed at: {}".format(version, path))
        return False
    return path


def _get_temp_folder():
    path = Path(tempfile.gettempdir()).joinpath("solcx-tmp-{}".format(os.getpid()))
    if path.exists():
        shutil.rmtree(str(path))
    path.mkdir()
    return path


def _download_solc(url, show_progress):
    if not show_progress:
        response = requests.get(url)
        content = response.content
    else:
        response = requests.get(url, stream=True)
        total_size = int(response.headers.get("content-length", 0))
        progress_bar = tqdm(total=total_size, unit="iB", unit_scale=True)
        content = bytes()

        for data in response.iter_content(1024, decode_unicode=True):
            progress_bar.update(len(data))
            content += data
        progress_bar.close()

    if response.status_code != 200:
        raise DownloadError(
            "Received status code {} when attempting to download from {}".format(
                response.status_code, url
            )
        )
    return content


def _install_solc_linux(version, show_progress, solcx_binary_path=None):
    download = DOWNLOAD_BASE.format(version, "solc-static-linux")
    binary_path = _check_for_installed_version(version, solcx_binary_path=solcx_binary_path)
    if binary_path:
        LOGGER.info("Downloading solc {} from {}".format(version, download))
        content = _download_solc(download, show_progress)
        with open(binary_path, "wb") as fp:
            fp.write(content)
        _chmod_plus_x(binary_path)


def _install_solc_windows(version, show_progress, solcx_binary_path=None):
    download = DOWNLOAD_BASE.format(version, "solidity-windows.zip")
    install_folder = _check_for_installed_version(version)
    if install_folder:
        temp_path = _get_temp_folder()
        content = _download_solc(download, show_progress)
        with zipfile.ZipFile(BytesIO(content)) as zf:
            zf.extractall(str(temp_path))
        install_folder = get_solc_folder(solcx_binary_path=solcx_binary_path).joinpath(
            "solc-" + version
        )
        temp_path.rename(install_folder)


def _install_solc_osx(version, allow_osx, show_progress, solcx_binary_path):
    if version.startswith("v0.4") and not allow_osx:
        raise ValueError(
            "Py-solc-x cannot build solc versions 0.4.x on OSX. If you install solc 0.4.x "
            "using brew and reload solcx, the installed version will be available. "
            "See https://github.com/ethereum/homebrew-ethereum for installation instructions.\n\n"
            "To ignore this error, include 'allow_osx=True' when calling solcx.install_solc()"
        )
    temp_path = _get_temp_folder()
    download = DOWNLOAD_BASE.format(version, "solidity_{}.tar.gz".format(version[1:]))
    binary_path = _check_for_installed_version(version)
    if not binary_path:
        return

    content = _download_solc(download, show_progress)
    with tarfile.open(fileobj=BytesIO(content)) as tar:
        tar.extractall(temp_path)
    temp_path = temp_path.joinpath("solidity_{}".format(version[1:]))

    try:
        _check_subprocess_call(
            ["sh", str(temp_path.joinpath("scripts/install_deps.sh"))],
            message="Running dependency installation script `install_deps.sh`",
        )
    except subprocess.CalledProcessError as e:
        LOGGER.warning(e, exc_info=True)

    original_path = os.getcwd()
    temp_path.joinpath("build").mkdir(exist_ok=True)
    os.chdir(str(temp_path.joinpath("build").resolve()))
    try:
        for cmd in (["cmake", ".."], ["make"]):
            _check_subprocess_call(cmd, message="Running {}".format(cmd[0]))
        temp_path.joinpath("build/solc/solc").rename(binary_path)
    except subprocess.CalledProcessError as e:
        raise OSError(
            "{} returned non-zero exit status {} while attempting to build solc from the source. "
            "This is likely due to a missing or incorrect version of an external dependency.\n\n"
            "You may be able to solve this by installing the specific version using brew: "
            "https://solidity.readthedocs.io/en/v0.6.0/installing-solidity.html#binary-packages"
            "".format(cmd[0], e.returncode)
        )
    finally:
        os.chdir(original_path)

    _chmod_plus_x(binary_path)


if __name__ == "__main__":
    argument_parser = argparse.ArgumentParser()
    argument_parser.add_argument("version")
    argument_parser.add_argument("--solcx-binary-path", default=None)
    args = argument_parser.parse_args()
    install_solc(args.version, solcx_binary_path=args.solcx_binary_path)



from __future__ import absolute_import

import functools
import json
import re

import semantic_version

from .exceptions import ContractsNotFound, SolcError
from .install import get_executable
from .utils.filesystem import is_executable_available
from .wrapper import solc_wrapper

VERSION_DEV_DATE_MANGLER_RE = re.compile(r"(\d{4})\.0?(\d{1,2})\.0?(\d{1,2})")
strip_zeroes_from_month_and_day = functools.partial(
    VERSION_DEV_DATE_MANGLER_RE.sub, r"\g<1>.\g<2>.\g<3>"
)


def is_solc_available():
    solc_binary = get_executable()
    return is_executable_available(solc_binary)


def get_solc_version_string(**kwargs):
    kwargs["version"] = True
    stdoutdata, stderrdata, command, proc = solc_wrapper(**kwargs)
    _, _, version_string = stdoutdata.partition("\n")
    if not version_string or not version_string.startswith("Version: "):
        raise SolcError(
            command=command,
            return_code=proc.returncode,
            stdin_data=None,
            stdout_data=stdoutdata,
            stderr_data=stderrdata,
            message="Unable to extract version string from command output",
        )
    return version_string.rstrip()


def get_solc_version(**kwargs):
    # semantic_version as of 2017-5-5 expects only one + to be used in string
    return semantic_version.Version(
        strip_zeroes_from_month_and_day(
            get_solc_version_string(**kwargs)[len("Version: ") :].replace("++", "pp")
        )
    )


def solc_supports_standard_json_interface(**kwargs):
    return get_solc_version() in semantic_version.SimpleSpec(">=0.4.11")


def _parse_compiler_output(stdoutdata):
    output = json.loads(stdoutdata)

    contracts = output.get("contracts", {})
    sources = output.get("sources", {})

    for path_str, data in contracts.items():
        if "abi" in data:
            data["abi"] = json.loads(data["abi"])
        key = path_str.rsplit(":", maxsplit=1)[0]
        if "AST" in sources.get(key, {}):
            data["ast"] = sources[key]["AST"]

    return contracts


ALL_OUTPUT_VALUES = (
    "abi",
    "asm",
    "ast",
    "bin",
    "bin-runtime",
    "clone-bin",
    "devdoc",
    "opcodes",
    "userdoc",
)


def compile_source(source, allow_empty=False, output_values=ALL_OUTPUT_VALUES, **kwargs):
    if "stdin" in kwargs:
        raise ValueError("The `stdin` keyword is not allowed in the `compile_source` function")
    if "combined_json" in kwargs:
        raise ValueError(
            "The `combined_json` keyword is not allowed in the `compile_source` function"
        )

    combined_json = ",".join(output_values)
    compiler_kwargs = dict(stdin=source, combined_json=combined_json, **kwargs)

    stdoutdata, stderrdata, command, proc = solc_wrapper(**compiler_kwargs)

    contracts = _parse_compiler_output(stdoutdata)

    if not contracts and not allow_empty:
        raise ContractsNotFound(
            command=command,
            return_code=proc.returncode,
            stdin_data=source,
            stdout_data=stdoutdata,
            stderr_data=stderrdata,
        )
    return contracts


def compile_files(source_files, allow_empty=False, output_values=ALL_OUTPUT_VALUES, **kwargs):
    if "combined_json" in kwargs:
        raise ValueError(
            "The `combined_json` keyword is not allowed in the `compile_files` function"
        )

    combined_json = ",".join(output_values)
    compiler_kwargs = dict(source_files=source_files, combined_json=combined_json, **kwargs)

    stdoutdata, stderrdata, command, proc = solc_wrapper(**compiler_kwargs)

    contracts = _parse_compiler_output(stdoutdata)

    if not contracts and not allow_empty:
        raise ContractsNotFound(
            command=command,
            return_code=proc.returncode,
            stdin_data=None,
            stdout_data=stdoutdata,
            stderr_data=stderrdata,
        )
    return contracts


def compile_standard(input_data, allow_empty=False, **kwargs):
    if not input_data.get("sources") and not allow_empty:
        raise ContractsNotFound(
            command=None,
            return_code=None,
            stdin_data=json.dumps(input_data, sort_keys=True, indent=2),
            stdout_data=None,
            stderr_data=None,
        )

    stdoutdata, stderrdata, command, proc = solc_wrapper(
        stdin=json.dumps(input_data), standard_json=True, **kwargs
    )

    compiler_output = json.loads(stdoutdata)
    if "errors" in compiler_output:
        has_errors = any(error["severity"] == "error" for error in compiler_output["errors"])
        if has_errors:
            error_message = "\n".join(
                tuple(
                    error["formattedMessage"]
                    for error in compiler_output["errors"]
                    if error["severity"] == "error"
                )
            )
            raise SolcError(
                command,
                proc.returncode,
                json.dumps(input_data),
                stdoutdata,
                stderrdata,
                message=error_message,
            )
    return compiler_output


def link_code(unlinked_bytecode, libraries):
    libraries_arg = ",".join(
        (":".join((lib_name, lib_address)) for lib_name, lib_address in libraries.items())
    )
    stdoutdata, stderrdata, _, _ = solc_wrapper(
        stdin=unlinked_bytecode, link=True, libraries=libraries_arg,
    )

    return stdoutdata.replace("Linking completed.", "").strip()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from setuptools import find_packages, setup

setup(
    name="py-solc-x",
    version="0.8.0",
    description="""Python wrapper around the solc binary with 0.5.x and 0.6.x support""",
    long_description_markdown_filename="README.md",
    author="Ben Hauser (forked from py-solc by Piper Merriam)",
    author_email="ben@hauser.id",
    url="https://github.com/iamdefinitelyahuman/py-solc-x",
    include_package_data=True,
    py_modules=["solcx"],
    setup_requires=["setuptools-markdown"],
    python_requires=">=3.4, <4",
    install_requires=["semantic_version>=2.8.1,<3", "requests>=2.19.0,<3"],
    license="MIT",
    zip_safe=False,
    keywords="ethereum solidity solc",
    packages=find_packages(exclude=["tests", "tests.*"]),
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
)



import os


def is_executable_available(program):
    def is_exe(fpath):
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

    fpath = os.path.dirname(program)
    if fpath:
        if is_exe(program):
            return True
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            path = path.strip('"')
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return True

    return False



from __future__ import absolute_import

from .install import (  # noqa: F401
    get_available_solc_versions,
    get_installed_solc_versions,
    get_solc_folder,
    import_installed_solc,
    install_solc,
    install_solc_pragma,
    set_solc_version,
    set_solc_version_pragma,
)
from .main import (  # noqa: F401
    compile_files,
    compile_source,
    compile_standard,
    get_solc_version,
    get_solc_version_string,
    link_code,
)

# check for installed version of solc
import_installed_solc()

# default to latest version
if get_installed_solc_versions():
    set_solc_version(get_installed_solc_versions()[-1], silent=True)
